package com.jtauber.pdf;

import java.util.Vector;
import java.io.IOException;
import java.io.PrintWriter;
							   
/**
 * PDF Document
 *
 * class representing a PDF document. The document is built up by calling
 * various methods and then finally output to given filehandle using output
 * method.
 * 
 * A PDF document consists of a series of numbered objects preceded by a
 * header and followed by an xref table and trailer. The xref table allows
 * for quick access to objects by listing their character positions within
 * the document. For this reason the PDF document must keep track of the
 * character position of each object.
 *
 * The document also keeps direct track of the /Root and /Info objects
 */ 
public class PDFDocument {
  
	private static final String pdfVersion = "1.2"; // the version of the PDF spec followed
	
	private int position;    // the current character position
	private Vector location; // the character position of each object
	private int objectcount; // the counter for object numbering
	private Vector objects;  // the objects themselves
	private int xref;
	private PDFRoot root;
	private PDFInfo info;
	private PDFResources resources;
	/**
	 * creates an empty PDF document
	 */
	public PDFDocument(String producer) {
		this.position = 0;
		this.location = new Vector();
		this.objectcount = 0;
		this.objects = new Vector();
		makeRoot();
		makeInfo(producer);
		makeResources();
	}
public PDFResources getResources() {
	return this.resources;
}
	/**
	 * make a Type1 /Font object for the given internal fontname, external 
	 * basefont name and encoding
	 */
	public PDFFont makeFont(String fontname, String basefont, String encoding) {
		PDFFont font = new PDFFont(++this.objectcount,fontname,basefont,encoding);
		this.objects.addElement(font);
		return font;
	}
	/**
	 * make an /Info object
	 */
	private void makeInfo(String producer) {
		this.info = new PDFInfo(++this.objectcount, producer);
		this.objects.addElement(this.info);
	}
	/**
	 * make a /Page object that uses the given resources object and the
	 * given stream as content
	 */
	public PDFPage makePage(PDFResources resources, PDFStream contents, int pagewidth, int pageheight) {
		PDFPage page = new PDFPage(++this.objectcount, resources, contents, pagewidth, pageheight);
		this.objects.addElement(page);
		this.root.addPage(page);
		return page;
	}
	/**
	 * make a resources object
	 */
	private void makeResources() {
		this.resources = new PDFResources(++this.objectcount);
		this.objects.addElement(this.resources);
	}
	/**
	 * make Root with the next number
	 */
	private void makeRoot() {
		this.root = new PDFRoot(++this.objectcount);
		this.objects.addElement(this.root);
		PDFPages rootPages = new PDFPages(++this.objectcount); // root of Pages hierarchy
		this.objects.addElement(rootPages);
		this.root.setRootPages(rootPages);
	}
	/**
	 * make a stream object with the next number
	 */
	public PDFStream makeStream() {
		this.objectcount = this.objectcount + 1;
		PDFStream obj = new PDFStream(this.objectcount);
		this.objects.addElement(obj);
		return obj;
	}
	/**
	 * write the entire document to the given writer, keeping track of
	 * character position along the way
	 */
	public void output(PrintWriter writer) throws IOException {
		this.position = this.position + outputHeader(writer);
		for (int i=1; i < this.objectcount+1; i++) {
			this.location.addElement(new Integer(this.position));
			PDFObject o = (PDFObject)this.objects.elementAt(i-1);
			this.position = this.position + o.output(writer);
		}
		this.position = this.position + outputXref(writer);
		outputTrailer(writer);
		writer.flush();
	}
	/**
	 * write the PDF header to the given writer
	 */
	private int outputHeader(PrintWriter writer) throws IOException {
		String pdf = "%PDF-" + this.pdfVersion + "\n";
		writer.write(pdf);
		return pdf.length();
	}
	/**
	 * output trailer to the given writer
	 */
	private void outputTrailer(PrintWriter writer) throws IOException {
		String pdf = "trailer\n<<\n";
		pdf = pdf + "/Size " + (this.objectcount+1) +"\n";
		pdf = pdf + "/Root " + this.root.number + " " + this.root.generation + " R\n";
		pdf = pdf + "/Info " + this.info.number + " " + this.info.generation + " R\n";
		pdf = pdf + ">>\nstartxref\n" + this.xref + "\n%%EOF\n";
		writer.write(pdf);
	}
	/**
	 * write the xref table to the given writer
	 */
	private int outputXref(PrintWriter writer) throws IOException {
		this.xref = this.position; // remember position for trailer
		String pdf = "xref\n0 " + (this.objectcount+1) + "\n0000000000 65535 f \n";
		for (int i=1; i < this.objectcount+1; i++) {
			String padding = "0000000000";
			String x = this.location.elementAt(i-1).toString();
			String loc = padding.substring(x.length()) + x;
			pdf = pdf + loc + " 00000 n \n";
		}
		writer.write(pdf);
		return pdf.length();
	}
}